/*
 * @file interfaces.h
 * @brief contains public interfaces to python objects (for C bindings)
 */
typedef struct {
        void (*renderFunc)(PyObject *self, int time, int picking);
        int (*compileFunc)(PyObject *self);
} IRender;

typedef struct {
        void (*appendFunc)(PyObject *self, int value);
        int (*itemFunc)(PyObject *self, int value, int *ret);
        int *(*intsFunc)(PyObject *self);
        int (*lengthFunc)(PyObject *self);
} IIntList;

typedef struct {
        PyObject *(*addFunc)(PyObject *a, PyObject *b);
        PyObject *(*multiplyScalarFunc)(PyObject *vector, double s);
        void (*translateByMeFunc)(PyObject *self);
        void (*rotateByMeFunc)(PyObject *self, double amt);
} IVector;

typedef struct {
        int (*lengthFunc)(PyObject *v);
        double *(*valuesFunc)(PyObject *v);
        void (*appendFunc)(PyObject *vlist, PyObject *vector);
        PyObject *(*itemFunc)(PyObject *self, int index);
} IVectorList;

typedef struct {
        void (*transformAndAdvanceFunc)(PyObject *self, PyObject *status,
                        int time);
        int (*registerSizeFunc)(PyObject *self);
	void *(*registerFunc)(PyObject *self, PyObject *r, 
                PyObject *args);
        void (*deRegisterFunc)(PyObject *self, PyObject *status);
        void (*invertFunc)(PyObject *self);
} IPath;

typedef struct {
        void (*generateTransform)(PyObject *self, PyObject *status, int time,
                        int gen_forward, int gen_reverse, int apply);
        void (*transformByMe)(PyObject *self);
        void (*invertByMe)(PyObject *self);
        int (*registerSize)(PyObject *self);
        void *(*registerFunc)(PyObject *self, PyObject *status, PyObject *args);
        void (*deRegisterFunc)(PyObject *self, PyObject *status);
} ITransform;


/*
typedef struct {
        void (*transformByMeFunc)(PyObject *self);
} ITransform;
*/

typedef struct {
        void (*setMaterial)(PyObject *self, int time);
        void (*unsetMaterial)(PyObject *self, int time);
} IMaterial;

typedef struct {
	void (*setTexture)(PyObject *self, int time);
} ITexture;

typedef struct {
	void (*run)(PyObject *self, double *v, double *n, int width,
		int height, int time);
} IDSBehaviour;

typedef struct {
	void (*Init)(PyObject *self, int time);
	void (*RenderStep)(PyObject *self, int time, int is_paused, 
				int paused_time);
	void (*Destroy)(PyObject *self, int time);
} IMixin;

typedef struct {
	int (*GetNodes)(PyObject *self, GraphNode ***nodes);
	int (*GetConnections)(PyObject *self, GraphConnection ***connections);
} IGraph;


typedef struct {
        IRender *renderInterface;
        IIntList *intListInterface;
        IVectorList *vectorListInterface;
        IPath *pathInterface;
        IVector *vectorInterface;
        ITransform *transformInterface;
        IMaterial *materialInterface;
	ITexture *textureInterface;
	IDSBehaviour *behaviourInterface;
	IMixin *mixinInterface;
	IGraph *graphInterface;
} IC;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
} TograMin;

#define GRAPH_IF(x)		(((TograMin *)x)->interfaces->graphInterface)
#define IS_GRAPH(x)		(GRAPH_IF(x) != NULL);
#define GET_NODES(x,c)		(GRAPH_IF(x)->GetNodes(x, c))
#define GET_CONNECTIONS(x,c)	(GRAPH_IF(x)->GetConnections(x, c))

#define RENDER_IF(x)            (((TograMin *)x)->interfaces->renderInterface)
#define IS_RENDERABLE(x)        (RENDER_IF(x) != NULL)
#define RENDER(x,t,p)       	(RENDER_IF(x)->renderFunc(x, t, p))
#define IS_STATIC(x)            (RENDER_IF(x)->isStaticFunc(x))
#define COMPILE(x)              (RENDER_IF(x)->compileFunc(x))

#define INTLIST_IF(x)           (((TograMin *)x)->interfaces->intListInterface)
#define IS_INTLIST(x)           (INTLIST_IF(x) != NULL)
#define IL_APPEND(x,v)          (INTLIST_IF(x)->appendFunc(x, v))
#define IL_ITEM(x,v,r)          (INTLIST_IF(x)->itemFunc(x, v, r))
#define IL_INTS(x)              (INTLIST_IF(x)->intsFunc(x))
#define IL_LENGTH(x)            (INTLIST_IF(x)->lengthFunc(x))

#define VECTOR_IF(x)            (((TograMin *)x)->interfaces->vectorInterface)
#define IS_VECTOR(x)            (VECTOR_IF(x) != NULL)
#define ADD(x, y)               (VECTOR_IF(x)->addFunc(x, y))
#define MULSCALAR(x, s)         (VECTOR_IF(x)->multiplyScalarFunc(x, s))
#define TRANSLATEBYME(x)        (VECTOR_IF(x)->translateByMeFunc(x))
#define ROTATEBYME(x,a)         (VECTOR_IF(x)->rotateByMeFunc(x, a))

#define VECTORLIST_IF(x)        (((TograMin *)x)->interfaces->\
                                                vectorListInterface)
#define IS_VECTORLIST(x)        (VECTORLIST_IF(x) != NULL)
#define VL_LENGTH(x)            (VECTORLIST_IF(x)->lengthFunc(x))
#define VL_VALUES(x)            (VECTORLIST_IF(x)->valuesFunc(x))
#define VL_APPEND(x, v)         (VECTORLIST_IF(x)->appendFunc(x, v))
#define VL_ITEM(x, n)           (VECTORLIST_IF(x)->itemFunc(x, n))

#define TRANSFORM_IF(x)         (((TograMin *)x)->interfaces->transformInterface)
#define IS_TRANSFORM(x)         (TRANSFORM_IF(x) != NULL)
#define GENERATE(s,u,t,f,r,a)   TRANSFORM_IF(s)->generateTransform(s,u,t,f,r,a)
#define TRANSFORMBYME(s)        TRANSFORM_IF(s)->transformByMe(s)
#define INVERTBYME(s)           TRANSFORM_IF(s)->invertByMe(s)
#define REGISTERSIZE(s)         TRANSFORM_IF(s)->registerSize(s)
#define REGISTER(s,st,a)        TRANSFORM_IF(s)->registerFunc(s,st,a)
#define DEREGISTER(s,st)        TRANSFORM_IF(s)->deRegisterFunc(s,st)

#define MATERIAL_IF(x)          (((TograMin *)x)->interfaces->materialInterface)
#define IS_MATERIAL(x)          (MATERIAL_IF(x) != NULL)
#define SET_MATERIAL(x,t)       (MATERIAL_IF(x)->setMaterial(x,t))
#define UNSET_MATERIAL(x,t)     (MATERIAL_IF(x)->unsetMaterial(x,t))

#define TEXTURE_IF(x)		(((TograMin *)x)->interfaces->textureInterface)
#define IS_TEXTURE(x)		(TEXTURE_IF(x) != NULL)
#define TEXTURE_SET(x,t)	(TEXTURE_IF(x)->setTexture(x,t))

#define DSBEHAVIOUR_IF(x)	(((TograMin *)x)->interfaces->behaviourInterface)
#define IS_DSBEHAVIOUR(x)	(DSBEHAVIOUR_IF(x) != NULL)
#define DSBEHAVIOUR_RUN(x,v,n,w,h,t)	(DSBEHAVIOUR_IF(x)->run(x,v,n,w,h,t))

#define MIXIN_IF(x)		(((TograMin *)x)->interfaces->mixinInterface)
#define IS_MIXIN(x)		(MIXIN_IF(x) != NULL)
#define INIT(x,t)		(MIXIN_IF(x)->Init(x,t))
#define RENDERSTEP(x,t,i,p)	(MIXIN_IF(x)->RenderStep(x,t,i,p))
#define DESTROY(x,t)		(MIXIN_IF(x)->Destroy(x,t))


#define TOGRA_DEFINE_MIXIN(type)					    \
staticforward PyTypeObject Py##type##Type;				    \
static void C_Py##type##_Init(PyObject *self, int time);		    \
static void C_Py##type##_RenderStep(PyObject *self, int time, 		    \
		int is_paused, int paused_time);			    \
static void C_Py##type##_Destroy(PyObject *self, int time);		    \
static IMixin Py##type##_IMixin = {C_Py##type##_Init, 			    \
	C_Py##type##_RenderStep, C_Py##type##_Destroy};		            \
static IC Py##type##_Interfaces = {NULL, NULL, NULL, NULL, NULL, NULL,	    \
	NULL, NULL, NULL, &Py##type##_IMixin};
	
#define TOGRA_DEFINE_RENDER(type)					    \
staticforward PyTypeObject Py##type##Type;				    \
static void C_Py##type##_Render(PyObject *self, int time, int picking);	    \
static int C_Py##type##_Compile(PyObject *self);  		            \
static IRender Py##type##_IRender = {C_Py##type##_Render, 		    \
	C_Py##type##_Compile};						    \
static IC Py##type##_Interfaces = {&Py##type##_IRender};

#define TOGRA_DEFINE_GRAPH(type)					    \
staticforward PyTypeObject Py##type##Type;				    \
static int C_Py##type##_GetNodes(PyObject *self, GraphNode ***nodes);	    \
static int C_Py##type##_GetConnections(PyObject *self,                      \
		       GraphConnection ***connections);  		    \
static IGraph Py##type##_IGraph = {C_Py##type##_GetNodes, 		    \
	C_Py##type##_GetConnections};					    \
static IC Py##type##_Interfaces = {NULL, NULL, NULL, NULL, NULL, NULL,      \
	NULL, NULL, NULL, NULL, &Py##type##_IGraph};


#define TOGRA_OBJECT_HEAD(t)		\
static PyTypeObject Py##t##Type = {	\
	PyObject_HEAD_INIT(NULL)	\
	0,				\
	"Py" #t,			\
	sizeof(Py##t),			\
	0,				\
	Py##t##_Free,

#define TOGRA_DEFINE_SIMPLE_OBJECT(t)	\
TOGRA_OBJECT_HEAD(t)			\
};

#define TOGRA_DEFINE_GETATTR_OBJECT(t)	\
TOGRA_OBJECT_HEAD(t)			\
	0,				\
	Py##t##_GetAttr,		\
};

/**
 * The PyVector type
 */
typedef struct {
        PyObject_HEAD
        IC *interfaces;
        double * vector;        /**< An array of length 3 */
        int memory_internal;    /**< TRUE if the system allocated the memory,
                                  FALSE if the vector was provided by the
                                  caller */
} PyVector;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        PyObject ** pyVectors;
        double * values;
        int length;
        int alloced;
} PyVectorList;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        PyObject * vectors;
        PyObject * normals;
        PyObject * vertexlist;
        int length;
        int mode;
} PyRender;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        PyObject * vectors;
        PyObject * normals;
        PyObject * vertexlist;
        PyObject * renderPath;
        PyObject * textureCoords;
        int textureMode;
        int length;
        int normal_mode;
        int display_list;
        int mode;
} PyStaticRender;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        PyObject * vectors;
        PyObject * normals;
        PyObject * renderPath;
        PyObject * textureCoords;
        int textureMode;
        int width;
	int height;
        int normal_mode;
        int display_list;
} PyStaticSurface;

struct PyDynamicSurface;

typedef struct PyDynamicSurface {
        PyObject_HEAD
        IC *interfaces;
        PyObject * vectors;
        PyObject * normals;
        PyObject * textureCoords;
	unsigned int *faces;
	double *vector_cache;
	double *normal_cache;
	PyObject *ds_behaviour;
	int textureMode;
        int width;
	int height;
} PyDynamicSurface;

typedef struct {
	PyObject_HEAD
	IC *interfaces;
	dWorldID world;
	dSpaceID space;
	dJointGroupID contactGroup;
	long lastTime;
} PyODEMixin; 

typedef struct {
	PyObject_HEAD
	IC *interfaces;
	PyObject *Init;
	PyObject *RenderStep;
	PyObject *Destroy;
} PyProtoMixin; 

typedef struct {
	PyObject_HEAD
	IC *interfaces;
	PyObject *Render;
	PyObject *Compile;
} PyProtoRender; 

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        int * ints;
        int length;
        int alloced;
} PyIntList;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        float * floats;
        int length;
        int alloced;
} PyFloatList;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        PyObject * pathPoints;
        PyObject * pathIndexes;
        void (*interpFunction)(PyObject *this, PyObject *a,                                     PyObject *b, int apos, int bpos, int cur, int gen_forward,
                        int gen_reverse, int apply);
        PyObject * inverseV;
        double inverseS;
        PyObject * forwardV;
        double forwardS;
        void (*invertFunction)(PyObject *this);
        void (*transformFunction)(PyObject *this);
        int maxIndex;
        int maxPos;
        int is_inited;
        int basePos;
} PyInterpolatedPath;

typedef struct {
        int curIndex;
        int is_called;
        int interpPointA;
        int interpPointB;
        int delay;
        int doCallback;
        PyObject *userPointer;
	int nextInterval;
	int intervalSize;
	int genIntervals;
	int wrap;
} PyIPStatus;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        PyObject **children;
        int length;
        int alloced;
        int is_static;
        int display_list;
        PyObject *transform;
        PyObject *renderPath;
        void *status;
        PyObject *material;
} PyContainer;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        PyObject **children;
        int length;
        int alloced;
        int is_static;
        int display_list;
        PyObject *transform;
        PyObject *renderPath;
        void *status;
        PyObject *material;
        int slices;
        int segs;
} PySphere;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        PyObject *reflectFace;
        PyObject *reflectTransform;
        PyObject *reflectedTree;
        PyObject *path;
        void *status;
} PyReflect;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        PyObject * translation;
        PyObject * rotationAxis;
        double rotationAmount;
        PyObject * scale;
} PyTransform;

typedef struct {
	PyObject_HEAD
	IC *interfaces;
	dBodyID body;
	dGeomID geom;
	PyODEMixin *mixin;
} PyODETransform;

typedef struct _C_PyRenderControl_LightObj _C_PyRenderControl_LightObj;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        PyObject *renderObject;
        long baseTime;
        long lastTime;
        long pausedTime;
        long pausedAmount;
        int isPaused;
        int isTimeBased;
        PyObject *view_from;
        PyObject *view_to;
        PyObject *view_up;
        PyObject *keyFunction;
        PyObject *mouseFunction;
        PyObject *motionFunction;
	PyObject *messageFunction;
        _C_PyRenderControl_LightObj *lights;
	TG_List *mixins;
} PyRenderControl;

struct PyMaterial;
typedef struct PyMaterial {
        PyObject_HEAD
        IC *interfaces;
        float ambient_colour[4];
        float emissive_colour[4];
        float specular_colour[4];
        float shininess;
        PyObject *texture;
	struct PyMaterial *old_colour;
} PyMaterial;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        int texture;
        int format;
} PyStaticTexture;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        double values[16];
        double inverse[16];
        int hasInverse;
} PyMatrix;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        PyObject **children;
        PyObject **status;
        int length;
        int alloced;
} PyTransformContainer;

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        int light;
        int transformWithTree;
} PyLightObject;

typedef struct {
	PyObject_HEAD
	IC *interfaces;
	double xper;
	double yper;
	double xmag;
	double ymag;
	double xperper;
	double yperper;
	double xpermag;
	double ypermag;
/*	int width;
	int height;
	*/
} PyWobbleBehaviour;

typedef struct {
	PyObject_HEAD
	IC *interfaces;
	PyObject *to_surface;
	int period;
} PyInterpolationBehaviour;	

typedef struct {
	PyObject_HEAD
	IC *interfaces;
	double xper;
	double yper;
	double xtimeper;
	double ytimeper;
} PyRippleBehaviour;	

typedef struct {
	int 		from;
	int 		to;
	double		length;
	double		strength;
} GraphMixinConnection;

typedef struct {
	PyObject_HEAD
	IC *interfaces;
	GraphMixinConnection *connections;
	unsigned int num_connections;
	PyObject **ODE_Objs;
	unsigned int num_objs;
	double friction;
	double repulsion;
} PyGraphMixin;

typedef struct {
	PyObject_HEAD
	IC *interfaces;
	PyObject *graph;
	double friction;
	double repulsion;
	double selectionForce;
	PyObject *selected;
	PyObject *destination;
	int distance;
	Partition_3D *partition;
} PyCustomGraphMixin;

typedef struct {
	PyObject_HEAD
	IC *interfaces;
	GraphNode **nodes;
	GraphConnection **connections;
	GraphGroup **groups;
	int num_nodes;
	int num_connections;
	int num_groups;
} PyGraph;
