/*
 * Copyright 2005, Shane Stephens.
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

/*
 * We will have:
 * 	(1) A list of ODE Transforms
 * 	(2) A list of connector data fields
 */

#include "TograIncludes.h"

#define TOGRA_PROPERTY_SET_GET(obj,pname,property,c_type,f_str,constructor)\
void C_Py##obj##_Set##pname(PyObject *self, c_type arg) {		\
	Py##obj *o = (Py##obj *)self;					\
	o->property = arg;						\
}									\
PyObject *Py##obj##_Set##pname(PyObject *self, PyObject *args) {	\
	c_type o;							\
	if (!PyArg_ParseTuple(args, f_str, &o))				\
		return NULL;						\
									\
	C_Py##obj##_Set##pname(self, o);				\
	Py_INCREF(Py_None);						\
	return Py_None;							\
}									\
c_type C_Py##obj##_Get##pname(PyObject *self) {				\
	Py##obj *o = (Py##obj *)self;					\
	return o->property;						\
}									\
PyObject *Py##obj##_Get##pname(PyObject *self, PyObject *args) {	\
	return constructor(C_Py##obj##_Get##pname(self));		\
}

TOGRA_DEFINE_MIXIN(CustomGraphMixin)

TOGRA_PROPERTY_SET_GET(CustomGraphMixin,Friction,friction,double,"d",
		PyFloat_FromDouble)
TOGRA_PROPERTY_SET_GET(CustomGraphMixin,Repulsion,repulsion,double,"d",
		PyFloat_FromDouble)

void *C_PyCustomGraphMixin_New(void) {

	PyCustomGraphMixin *om = PyObject_New(PyCustomGraphMixin, &PyCustomGraphMixinType);
	D_ALLOC(om, "CustomGraphMixin");

	om->interfaces = &PyCustomGraphMixin_Interfaces;

	om->graph = NULL;

	om->friction = 20;
	om->repulsion = 500;

	om->distance = 20;
	om->partition = NULL;

	om->selected = NULL;
	om->destination = NULL;
	om->selectionForce = 500;
	
	return (void *)om;

}

PyObject *PyCustomGraphMixin_New(PyObject *self, PyObject *args) {

	return C_PyCustomGraphMixin_New();

}

void C_PyCustomGraphMixin_Init(PyObject *self, int time) {
	
	return;
	
}

void C_PyCustomGraphMixin_Destroy(PyObject *self, int time) {
	
	return;
}

static void PyCustomGraphMixin_Free(PyObject *self) {

	D_FREE(self, "CustomGraphMixin");

}

PyObject *C_PyODETransform_GetPosition(PyObject *p);
double *PyVector_Vector(PyObject *p);
void C_PyODETransform_AddForce(PyObject *p, double x, double y, double z);
PyObject *C_PyODETransform_GetVelocity(PyObject *p);

typedef struct {
	PyObject **nodes;
	int length;
} GraphNodeList;

GraphNodeList *graphNodeList_create(void) {

	GraphNodeList *gnl = malloc(sizeof(GraphNodeList));
	memset(gnl, 0, sizeof(GraphNodeList));
	return gnl;
}

void graphNodeList_append(GraphNodeList *gnl, PyObject *node) {

	gnl->length += 1;
	gnl->nodes = realloc(gnl->nodes, sizeof(PyObject *) * gnl->length);
	gnl->nodes[gnl->length - 1] = node;

}
	
int graphNodeList_remove(GraphNodeList *gnl, int pos) {

	memcpy(gnl->nodes + pos, gnl->nodes + pos + 1, 
		(gnl->length - pos - 1) * sizeof(PyObject *));
	gnl->length -= 1;
	if (gnl->length == 0) {
		free(gnl->nodes);
		return 1;
	}	
	gnl->nodes = realloc(gnl->nodes, gnl->length * sizeof(PyObject *));
	return 0;

}

void graphNodeList_add(GraphNodeList *gnl, GraphNodeList *other) {

	gnl->nodes = realloc(gnl->nodes, 
		(gnl->length + other->length) * sizeof(PyObject *));
	memcpy(gnl->nodes + gnl->length, other->nodes, 
		other->length * sizeof(PyObject *));
	gnl->length += other->length;

}

void partition_3d_unfixkey(int *x, int *y, int *z);
int _cgm_do_forces_func(int x, int y, int z, GraphNodeList *gnl, 
		PyCustomGraphMixin *om) {

	int j, k, l;
	GraphNodeList *effectees;
		
	effectees = graphNodeList_create();
		
	for (j = -1; j < 2; j++) {
		for (k = -1; k < 2; k++) {
			for (l = -1; l < 2; l++) {
				GraphNodeList *gnl_t;

				gnl_t = partition_3d_retrieve(om->partition,
					x + j, y + k, z + l);

				if (gnl_t == NULL)
					continue;

				graphNodeList_add(effectees, gnl_t);
			}
		}
	}

	for (j = 0; j < gnl->length; j++) {
		PyObject *f_node = gnl->nodes[j];
		PyObject *p1 = C_PyODETransform_GetPosition(f_node);
		double *v1 = PyVector_Vector(p1);
		PyObject *v;
		double *vv;
		
		for (k = 0; k < effectees->length; k++) {
			PyObject *t_node = effectees->nodes[k];
			PyObject *p2;
			if (t_node == f_node)
				continue;
			p2 = C_PyODETransform_GetPosition(t_node);
			double *v2 = PyVector_Vector(p2);
			double pd[3];
			double d2;
			double ff;

			for(l = 0; l < 3; l++)
				pd[l] = v1[l] - v2[l];

			d2 = pd[0] * pd[0] + pd[1] * pd[1] + 
				pd[2] * pd[2];
			if (d2 > om->distance * om->distance)
				continue;

			ff = om->repulsion / d2;

			C_PyODETransform_AddForce(f_node,
				pd[0] * ff, pd[1] * ff, pd[2] * ff);
			C_PyODETransform_AddForce(t_node,
				-pd[0] * ff, -pd[1] * ff, -pd[2] * ff);

			Py_DECREF(p2);
		
		}
		
		Py_DECREF(p1);
		
		v = C_PyODETransform_GetVelocity(f_node);
		vv = PyVector_Vector(v);
	
		C_PyODETransform_AddForce(f_node,
			-vv[0] * om->friction, -vv[1] * om->friction, 
			-vv[2] * om->friction);

		Py_DECREF(v);


	}	

	return 0;

}
	
int _cgm_update(int x, int y, int z, GraphNodeList *gnl, 
		PyCustomGraphMixin *om) {

	int j, rval = 0;
	
	int l = gnl->length;

	for (j = 0; j < l; j++) {
		
		GraphNodeList *gnl_t;
		PyObject *node = gnl->nodes[j];
		PyObject *pos = C_PyODETransform_GetPosition(node);
		double *pos_v = PyVector_Vector(pos);
		int xx, yy, zz;
		
		xx = ((int)pos_v[0]) / om->distance;
		yy = ((int)pos_v[1]) / om->distance;
		zz = ((int)pos_v[2]) / om->distance;

		if ((xx == x) && (yy == y) && (zz == z)) {
			continue;
		}

		if(graphNodeList_remove(gnl, j)) {
			free(gnl);
			rval = 1;
		}
		
		j -= 1;
		l -= 1;
	
		gnl_t = partition_3d_retrieve(om->partition, xx, yy, zz);
		if (gnl_t == NULL) {
			gnl_t = graphNodeList_create();
			partition_3d_insert(om->partition, xx, yy, zz, gnl_t);
		}
	
		graphNodeList_append(gnl_t, node);
	}


	return rval;

}

void C_PyCustomGraphMixin_RenderStep(PyObject *self, int time, int is_paused,
		int pausedTime) {
	PyCustomGraphMixin *om = (PyCustomGraphMixin *)self;

	int i, j, k;

	GraphNode **nodes;
	GraphConnection **connections;

	int num_connections, num_nodes;

	num_nodes = GET_NODES(om->graph, &nodes);
	num_connections = GET_CONNECTIONS(om->graph, &connections);

	if (om->selected != NULL) {

		int l = PyList_Size(om->selected);
		
		if (l > PyList_Size(om->destination)) {
			l = PyList_Size(om->destination);
		}

		for (i = 0; i < l; i++) {
			PyObject *item = PyList_GetItem(om->selected, i);
			PyObject *p = C_PyODETransform_GetPosition(item);
			PyObject *pos = PyList_GetItem(om->destination, i);
			double *v1 = PyVector_Vector(p);
			double *v2 = PyVector_Vector(pos);
			double pd[3];
			double ff = om->selectionForce;

			for (j = 0; j < 3; j++)
				pd[j] = v2[j] - v1[j];

			C_PyODETransform_AddForce(item,
					ff * pd[0], ff * pd[1], ff * pd[2]);

			Py_DECREF(p);
		}
	}	
		
	for (i = 0; i < num_connections; i++) {
		
		GraphConnection *cur = connections[i];
		PyObject *p1 = C_PyODETransform_GetPosition(cur->from);
		PyObject *p2 = C_PyODETransform_GetPosition(cur->to);
		double *v1 = PyVector_Vector(p1);
		double *v2 = PyVector_Vector(p2);
		double pd[3];
		double d2;
		double d;
		double ff;
		
		for (j = 0; j < 3; j++)
			pd[j] = v1[j] - v2[j];
		
		d2 = pd[0]*pd[0] + pd[1]*pd[1] + pd[2]*pd[2];
		d = sqrt(d2) - cur->length;

		ff = d * cur->strength;

		C_PyODETransform_AddForce(cur->to,
				pd[0] * ff, pd[1] * ff, pd[2] * ff);
		C_PyODETransform_AddForce(cur->from,
				-pd[0] * ff, -pd[1] * ff, -pd[2] * ff);

		Py_DECREF(p1);
		Py_DECREF(p2);
	}
	
	if (om->distance > 0)
	{
		if (om->partition == NULL)
		{

			om->partition = create_partition_3d();
			
			for (i = 0; i < num_nodes; i++)
			{
				int x, y, z;
				PyObject *pos = C_PyODETransform_GetPosition(
						(nodes[i])->object);
				double *pos_v = PyVector_Vector(pos);
				GraphNodeList *gnl;
				
				x = ((int)pos_v[0]) / om->distance;
				y = ((int)pos_v[1]) / om->distance;
				z = ((int)pos_v[2]) / om->distance;

				gnl = partition_3d_retrieve(om->partition,
						x, y, z);
				if (gnl == NULL) {
					gnl = graphNodeList_create();
					partition_3d_insert(om->partition,
							x, y, z, gnl);
				}

				graphNodeList_append(gnl, (nodes[i])->object);
				
			}
		}
		else
		{
			partition_3d_foreach(om->partition,
				(partition_3d_foreach_func)_cgm_update, om);
		}

		partition_3d_foreach(om->partition, 
			(partition_3d_foreach_func)_cgm_do_forces_func, om);

		return;
	}

	for (i = 0; i < num_nodes; i++)
	{
		PyObject *v;
		double *vv;
		PyObject *p1 = C_PyODETransform_GetPosition(
				(nodes[i])->object);
		double *v1 = PyVector_Vector(p1);
		
		for (j = i + 1; j < num_nodes; j++)
		{
			PyObject *p2 = C_PyODETransform_GetPosition(
					nodes[j]->object);
			double *v2 = PyVector_Vector(p2);
			double pd[3];
			double d2;
			double ff;

			for(k = 0; k < 3; k++)
				pd[k] = v1[k] - v2[k];

			d2 = pd[0] * pd[0] + pd[1] * pd[1] + pd[2] * pd[2];
			ff = om->repulsion / d2;

			C_PyODETransform_AddForce(nodes[i]->object,
					pd[0] * ff, pd[1] * ff, pd[2] * ff);
			C_PyODETransform_AddForce(nodes[j]->object,
					-pd[0] * ff, -pd[1] * ff, -pd[2] * ff);

			Py_DECREF(p2);
			
		}

		Py_DECREF(p1);
		
		v = C_PyODETransform_GetVelocity(nodes[i]->object);
		vv = PyVector_Vector(v);
		
		C_PyODETransform_AddForce(nodes[i]->object,
				-vv[0] * om->friction, -vv[1] * om->friction, 
				-vv[2] * om->friction);
	}
		
		
}

void C_PyCustomGraphMixin_SetGraph(PyObject *self, PyObject *graph) {

	PyCustomGraphMixin *cgm = (PyCustomGraphMixin *)self;

	if (cgm->graph != NULL) {
		Py_DECREF(cgm->graph);
	}
	
	cgm->graph = graph;
	
	Py_INCREF(graph);
}	

PyObject *PyCustomGraphMixin_SetGraph(PyObject *self, PyObject *args) {

	PyObject *graph;
	
	if (!PyArg_ParseTuple(args, "O", &graph)) {
		return NULL;
	}

	C_PyCustomGraphMixin_SetGraph(self, graph);

	Py_INCREF(Py_None);
	return Py_None;

}

void C_PyCustomGraphMixin_SetSelected(PyObject *self, PyObject *selected,
		PyObject *destination) {

	PyCustomGraphMixin *cgm = (PyCustomGraphMixin *)self;

	if (cgm->selected != NULL) {
		Py_DECREF(cgm->selected);
		Py_DECREF(cgm->destination);
	}

	if (selected == Py_None) {
		cgm->selected = NULL;
		cgm->destination = NULL;
		return;
	}
	
	cgm->selected = selected;
	Py_INCREF(selected);
	
	cgm->destination = destination;
	Py_INCREF(destination);
}
	
PyObject *PyCustomGraphMixin_SetSelected(PyObject *self, PyObject *args) {

	PyObject *selected;
	PyObject *destination;
	
	if (!PyArg_ParseTuple(args, "OO", &selected, &destination)) {
		return NULL;
	}

	C_PyCustomGraphMixin_SetSelected(self, selected, destination);

	Py_INCREF(Py_None);
	return Py_None;
}

static PyMethodDef PyCustomGraphMixin_methods[] = {
	{"setFriction", (PyCFunction)PyCustomGraphMixin_SetFriction, 
		METH_VARARGS, "set the friction of this graph object"},
	{"setRepulsion", (PyCFunction)PyCustomGraphMixin_SetRepulsion, 
		METH_VARARGS, "set the repulsion of this repulsion object"},
	{"setGraph", (PyCFunction)PyCustomGraphMixin_SetGraph,
		METH_VARARGS, "set the graph model for this object"},
	{"setSelected", (PyCFunction)PyCustomGraphMixin_SetSelected,
		METH_VARARGS, "set the selected node for this object"},
	{NULL, NULL, 0, NULL}
};

static PyObject *PyCustomGraphMixin_GetAttr(PyObject *self, char *name) {
	return Py_FindMethod(PyCustomGraphMixin_methods, self, name);
}

TOGRA_DEFINE_GETATTR_OBJECT(CustomGraphMixin)

