/*
 * Copyright 2003, Shane Stephens.
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

/*
 * We will have:
 * 	(1) A list of ODE Transforms
 * 	(2) A list of connector data fields
 */

#include "TograIncludes.h"

#define TOGRA_PROPERTY_SET_GET(obj,pname,property,c_type,f_str,constructor)\
void C_Py##obj##_Set##pname(PyObject *self, c_type arg) {		\
	Py##obj *o = (Py##obj *)self;					\
	o->property = arg;						\
}									\
PyObject *Py##obj##_Set##pname(PyObject *self, PyObject *args) {	\
	c_type o;							\
	if (!PyArg_ParseTuple(args, f_str, &o))				\
		return NULL;						\
									\
	C_Py##obj##_Set##pname(self, o);				\
	Py_INCREF(Py_None);						\
	return Py_None;							\
}									\
c_type C_Py##obj##_Get##pname(PyObject *self) {				\
	Py##obj *o = (Py##obj *)self;					\
	return o->property;						\
}									\
PyObject *Py##obj##_Get##pname(PyObject *self, PyObject *args) {	\
	return constructor(C_Py##obj##_Get##pname(self));		\
}

TOGRA_DEFINE_MIXIN(GraphMixin)

TOGRA_PROPERTY_SET_GET(GraphMixin,Friction,friction,double,"d",
		PyFloat_FromDouble)
TOGRA_PROPERTY_SET_GET(GraphMixin,Repulsion,repulsion,double,"d",
		PyFloat_FromDouble)

void *C_PyGraphMixin_New(void) {

	PyGraphMixin *om = PyObject_New(PyGraphMixin, &PyGraphMixinType);
	D_ALLOC(om, "GraphMixin");

	om->interfaces = &PyGraphMixin_Interfaces;

	om->connections = NULL;
	om->num_connections = 0;

	om->ODE_Objs = NULL;
	om->num_objs = 0;

	om->friction = 20;
	om->repulsion = 500;
	
	return (void *)om;

}

PyObject *PyGraphMixin_New(PyObject *self, PyObject *args) {

	return C_PyGraphMixin_New();

}

void C_PyGraphMixin_Init(PyObject *self, int time) {
	
	return;
	
}

void C_PyGraphMixin_Destroy(PyObject *self, int time) {
	
	return;
}

static void PyGraphMixin_Free(PyObject *self) {

	D_FREE(self, "GraphMixin");

}

PyObject *C_PyODETransform_GetPosition(PyObject *p);
double *PyVector_Vector(PyObject *p);
void C_PyODETransform_AddForce(PyObject *p, double x, double y, double z);
PyObject *C_PyODETransform_GetVelocity(PyObject *p);

void C_PyGraphMixin_RenderStep(PyObject *self, int time, int is_paused,
		int pausedTime) {
	PyGraphMixin *om = (PyGraphMixin *)self;

	int i, j, k;

	for (i = 0; i < om->num_connections; i++)
	{
		GraphMixinConnection *cur = om->connections + i;
		PyObject *o1 = om->ODE_Objs[cur->from];
		PyObject *o2 = om->ODE_Objs[cur->to];
		PyObject *p1 = C_PyODETransform_GetPosition(o1);
		PyObject *p2 = C_PyODETransform_GetPosition(o2);
		double *v1 = PyVector_Vector(p1);
		double *v2 = PyVector_Vector(p2);
		double pd[3];
		double d2;
		double d;
		double ff;
		
		for (j = 0; j < 3; j++)
			pd[j] = v1[j] - v2[j];
		
		d2 = pd[0]*pd[0] + pd[1]*pd[1] + pd[2]*pd[2];
		d = sqrt(d2) - cur->length;

		ff = d * cur->strength;

		C_PyODETransform_AddForce(o2,
				pd[0] * ff, pd[1] * ff, pd[2] * ff);
		C_PyODETransform_AddForce(o1,
				-pd[0] * ff, -pd[1] * ff, -pd[2] * ff);

		Py_DECREF(p1);
		Py_DECREF(p2);
	}

	for (i = 0; i < om->num_objs; i++)
	{
		PyObject *v;
		double *vv;
		PyObject *p1 = C_PyODETransform_GetPosition(
				om->ODE_Objs[i]);
		
		for (j = i + 1; j < om->num_objs; j++)
		{
			PyObject *p2 = C_PyODETransform_GetPosition(
					om->ODE_Objs[j]);
			double *v1 = PyVector_Vector(p1);
			double *v2 = PyVector_Vector(p2);
			double pd[3];
			double d2;
			double ff;

			for(k = 0; k < 3; k++)
				pd[k] = v1[k] - v2[k];

			d2 = pd[0] * pd[0] + pd[1] * pd[1] + pd[2] * pd[2];
			ff = om->repulsion / d2;

			C_PyODETransform_AddForce(om->ODE_Objs[i],
					pd[0] * ff, pd[1] * ff, pd[2] * ff);
			C_PyODETransform_AddForce(om->ODE_Objs[j],
					-pd[0] * ff, -pd[1] * ff, -pd[2] * ff);

			Py_DECREF(p2);
			
		}

		Py_DECREF(p1);
		
		v = C_PyODETransform_GetVelocity(om->ODE_Objs[i]);
		vv = PyVector_Vector(v);
		
		C_PyODETransform_AddForce(om->ODE_Objs[i], 
				-vv[0] * om->friction, -vv[1] * om->friction, 
				-vv[2] * om->friction);
	}
		
		
}

static int C_PyGraphMixin_AppendConnection(PyObject *self, int from, int to, 
		double length, double strength) {

	PyGraphMixin *gm = (PyGraphMixin *)self;

	gm->num_connections++;
	gm->connections = realloc(gm->connections, 
			sizeof(GraphMixinConnection) * gm->num_connections);

	gm->connections[gm->num_connections -1].from = from;
	gm->connections[gm->num_connections -1].to = to;
	gm->connections[gm->num_connections -1].length = length;
	gm->connections[gm->num_connections -1].strength = strength;

	return gm->num_connections - 1;
}
	
PyObject *PyGraphMixin_AppendConnection(PyObject *self, PyObject *args)
{
	int from, to;
	double length, strength;

	if (!PyArg_ParseTuple(args, "iidd", &from, &to, &length, &strength))
		return NULL;

	return PyInt_FromLong((long)C_PyGraphMixin_AppendConnection(self,
				from, to, length, strength));
}

static int C_PyGraphMixin_AppendObject(PyObject *self, PyObject *obj) {

	PyGraphMixin *gm = (PyGraphMixin *)self;

	gm->num_objs++;
	gm->ODE_Objs = realloc(gm->ODE_Objs, sizeof(PyObject *) * gm->num_objs);
	gm->ODE_Objs[gm->num_objs - 1] = obj;

	return gm->num_objs - 1;

}

PyObject *PyGraphMixin_AppendObject(PyObject *self, PyObject *args) {

	PyObject *o;

	if (!PyArg_ParseTuple(args, "O", &o))
		return NULL;

	Py_INCREF(o);

	return PyInt_FromLong((long)C_PyGraphMixin_AppendObject(self, o));
}

void C_PyGraphMixin_RemoveObjectConnections(PyObject *self, 
		int index) {
	
	int i;
	PyGraphMixin *gm = (PyGraphMixin *)self;
	int changed = 0;
	
	for (i = 0; i < gm->num_connections; i++)
	{
		GraphMixinConnection *c = gm->connections + i;

		if ((c->from == index) || (c->to == index)) {
			memcpy(gm->connections + i, gm->connections + i + 1,
				(gm->num_connections - i - 1) * 
				sizeof(GraphMixinConnection));
			gm->num_connections -= 1;
			changed = 1;
			i -= 1;
		}
		
	}

	if (changed)
		gm->connections = realloc(gm->connections,
			gm->num_connections * sizeof(GraphMixinConnection));
	
}

PyObject *PyGraphMixin_RemoveObjectConnections(PyObject *self, PyObject *args) {

	int i;
	
	if (!PyArg_ParseTuple(args, "i", &i))
		return NULL;

	C_PyGraphMixin_RemoveObjectConnections(self, i);

	Py_INCREF(Py_None);
	return Py_None;
}

static PyMethodDef PyGraphMixin_methods[] = {
	{"setFriction", (PyCFunction)PyGraphMixin_SetFriction, METH_VARARGS,
		"set the friction of this graph object"},
	{"setRepulsion", (PyCFunction)PyGraphMixin_SetRepulsion, METH_VARARGS,
		"set the repulsion of this repulsion object"},
	{"appendConnection", (PyCFunction)PyGraphMixin_AppendConnection,
		METH_VARARGS, "add a connection to this object"},
	{"appendObject", (PyCFunction)PyGraphMixin_AppendObject,
		METH_VARARGS, "add an object to this object"},
	{"removeObjectConnections", 
		(PyCFunction)PyGraphMixin_RemoveObjectConnections,
		METH_VARARGS, "remove the connections of an object"},
	{NULL, NULL, 0, NULL}
};

static PyObject *PyGraphMixin_GetAttr(PyObject *self, char *name) {
	return Py_FindMethod(PyGraphMixin_methods, self, name);
}

TOGRA_DEFINE_GETATTR_OBJECT(GraphMixin)

