/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"
staticforward PyTypeObject PyInterpolatedPathType;

void LinearPathInterp(PyObject *this, PyObject *a, PyObject *b, 						int apos, int bpos, int cur, int gen_forward,
                                int gen_reverse, int apply);

void BezierPathInterp(PyObject *this, PyObject *a, PyObject *b, int apos,
 				int bpos, int cur, int gen_forward,
                                int gen_reverse, int apply);
void LinearRotateInterp(PyObject *this, PyObject *a, PyObject *b,
				int apos, int bpos, int cur, int gen_forward,
                                int gen_reverse, int apply);

void PathInvert(PyObject *this);
void RotateInvert(PyObject *this);
void PathApply(PyObject *this);
void RotateApply(PyObject *this);

static void C_PyInterpolatedPath_GenerateTransform(PyObject *self, 
                PyObject *status, int time, int gen_forward, int gen_reverse,
                int apply);
static void C_PyInterpolatedPath_TransformByMe(PyObject *self);
static void C_PyInterpolatedPath_InvertByMe(PyObject *self);
static int C_PyInterpolatedPath_RegisterSize(PyObject *self);
static void *C_PyInterpolatedPath_Register(PyObject *self, PyObject *r, 
                PyObject *args);
static void C_PyInterpolatedPath_DeRegister(PyObject *self, PyObject *status); 

static ITransform PyInterpolatedPath_IPath = 
        {C_PyInterpolatedPath_GenerateTransform, 
        C_PyInterpolatedPath_TransformByMe, C_PyInterpolatedPath_InvertByMe,
        C_PyInterpolatedPath_RegisterSize, C_PyInterpolatedPath_Register,
        C_PyInterpolatedPath_DeRegister};
static IC PyInterpolatedPath_Interfaces = 
        {NULL, NULL, NULL, NULL, NULL, &PyInterpolatedPath_IPath};

static inline int PyInterpolatedPath_Check(PyObject *v) {

	return v->ob_type == &PyInterpolatedPathType;

}

static inline PyObject * PyInterpolatedPath_PathPoints(PyObject *v) {

    return ((PyInterpolatedPath *)v)->pathPoints;

}

static inline PyObject * PyInterpolatedPath_PathIndexes(PyObject *v) {

    return ((PyInterpolatedPath *)v)->pathIndexes;

}

static inline int PyInterpolatedPath_MaxIndex(PyObject *v) {

    return ((PyInterpolatedPath *)v)->maxIndex;

}

PyObject *C_PyInterpolatedPath_New(int isTimeBased) {
        PyInterpolatedPath *ip = PyObject_New(PyInterpolatedPath, 
                        &PyInterpolatedPathType);

        D_ALLOC(ip, "InterpolatedPath");
        
        ip->interfaces = &PyInterpolatedPath_Interfaces;
        ip->pathPoints = (PyObject *)PyList_New(0);
        ip->pathIndexes = (PyObject *)C_PyIntList_New();
        ip->maxIndex = 0;
        ip->interpFunction = BezierPathInterp;
        ip->invertFunction = PathInvert;
        ip->transformFunction = PathApply;
        ip->is_inited = 0;
        ip->inverseV = NULL;
        ip->inverseS = 0;
        Py_INCREF(ip->pathIndexes);
        Py_INCREF(ip->pathPoints);

        return (PyObject *)ip;
}

PyObject *C_PyInterpolatedPath_OfLength(int length, int isTimeBased) {
	PyInterpolatedPath *ip = 
                (PyInterpolatedPath *)C_PyInterpolatedPath_New(isTimeBased);
        Py_DECREF(ip->pathIndexes);
        ip->pathIndexes = (PyObject *)C_PyIntList_OfLength(length);
        Py_INCREF(ip->pathIndexes);
        
	return (PyObject *)ip;

}

PyObject *PyInterpolatedPath_New(PyObject *self, PyObject *args) {

	int length = 0;
        
	PyArg_ParseTuple(args, "|i", &length);
	
	if (length > 0) {	
                return (PyObject *)C_PyInterpolatedPath_OfLength(length, FALSE);
	}
	
    return (PyObject *)C_PyInterpolatedPath_New(FALSE);
    
}

static void C_PyInterpolatedPath_Free(void *self) {

        PyInterpolatedPath *ip = (PyInterpolatedPath *)self;

        D_FREE(self, "InterpolatedPath");
    
        Py_DECREF(ip->pathPoints);
        Py_DECREF(ip->pathIndexes);

        PyObject_Del((PyObject *)ip);
}
        
static void PyInterpolatedPath_Free(PyObject *self) {

    C_PyInterpolatedPath_Free(self);
        
}

void C_PyInterpolatedPath_Append(PyObject *self, PyObject *point, int index) {

        PyInterpolatedPath *ip = (PyInterpolatedPath *)self;

        Py_INCREF(point);
        
        PyList_Append(ip->pathPoints, point);
	IL_APPEND(ip->pathIndexes, index);
	
	ip->maxIndex++;
	
}

static PyObject *PyInterpolatedPath_Append(PyObject *self, PyObject *args) {

    PyObject *point, *index;
        
	if (!PyArg_ParseTuple(args, "OO!", &point, &PyInt_Type,	&index)) {
		return NULL;
	}
		
        C_PyInterpolatedPath_Append((void *)self, (void *)point, 
                        (int)PyInt_AsLong(index));
        
	Py_INCREF(Py_None);
	return Py_None;

}

void LinearPathInterp(PyObject *this, PyObject *a, PyObject *b, 						int apos, int bpos, int cur, int gen_forward,
                                int gen_reverse, int apply) {

	PyObject *accum;
	int diff1, diff2;
	PyInterpolatedPath *ip = (PyInterpolatedPath *)this;

	diff1 = cur - apos;
	diff2 = bpos - cur;
	
	accum = MULSCALAR(a, diff2);
	accum = ADD(accum, MULSCALAR(b, diff1));
	ip->forwardV = MULSCALAR(accum, 1.0 / (diff1 + diff2));
        
        if (apply == 1)     
                TRANSLATEBYME(ip->forwardV);

        if (gen_reverse)
        {
        	ip->inverseV = MULSCALAR(ip->forwardV, -1.0);
                if (apply == -1)
                        TRANSLATEBYME(ip->inverseV);
        }
	

}

	

void LinearRotateInterp(PyObject *this, PyObject *a, PyObject *b,
				int apos, int bpos, int cur, int gen_forward,
                                int gen_reverse, int apply) {
	
	PyObject *accum, *a1, *b1;
	double a2, b2, r;
	int diff1, diff2;
	PyInterpolatedPath *ip = (PyInterpolatedPath *)this;
        
	a1 = PyTuple_GetItem(a, 0);
	b1 = PyTuple_GetItem(b, 0);
	
	a2 = PyFloat_AsDouble(PyTuple_GetItem(a, 1));
	b2 = PyFloat_AsDouble(PyTuple_GetItem(b, 1));
	
	diff1 = cur - apos;
	diff2 = bpos - cur;

	accum = MULSCALAR(a1, diff2);
	accum = ADD(accum, MULSCALAR(b1, diff1));
        ip->forwardV = MULSCALAR(accum, 1.0 / (diff1 + diff2));
	
	r = (a2 * diff2 + b2 * diff1) / (diff1 + diff2);
        ip->forwardS = r;

        if (apply == 1)
        	ROTATEBYME(ip->forwardV, r);

        if (gen_reverse)
        {
                ip->inverseV = ip->forwardV;
                ip->inverseS = -r;
                if (apply == -1)
                        ROTATEBYME(ip->forwardV, -r);
        }
	
}

void BezierPathInterp(PyObject *this, PyObject *a, PyObject *b, int apos,
 				int bpos, int cur, int gen_forward,
                                int gen_reverse, int apply) {
	
	double index = (cur - apos) / ((double)(bpos - apos));
	double i1, i2, i3, i4;
	PyInterpolatedPath *ip = (PyInterpolatedPath *)this;
	
	PyObject *v1, *v2, *v3, *v4, *accum, *o1, *o2;
	
	i1 = index * index * index;
	i2 = 3 * index * index * (1 - index);
	i3 = 3 * index * (1 - index) * (1 - index);
	i4 = (1 - index) * (1 - index) * (1 - index);
	
	v1 = PyTuple_GetItem(a, 0);
	o1 = PyTuple_GetItem(a, 2);
	v4 = PyTuple_GetItem(b, 0);
	o2 = PyTuple_GetItem(b, 1);
	
	v2 = ADD(v1, o1);
	v3 = ADD(v4, o2);
	
	accum = C_PyVector_New();
	accum = ADD(accum, MULSCALAR(v1, i4));
	accum = ADD(accum, MULSCALAR(v2, i3));
	accum = ADD(accum, MULSCALAR(v3, i2));
	ip->forwardV = ADD(accum, MULSCALAR(v4, i1));

        if (apply == 1)
        	TRANSLATEBYME(ip->forwardV);

        if (gen_reverse)
        {
        	ip->inverseV = MULSCALAR(ip->forwardV, -1.0);
                if (apply == -1)
                        TRANSLATEBYME(ip->inverseV);
        }

}

void PathInvert(PyObject *this)
{

	PyInterpolatedPath *ip = (PyInterpolatedPath *)this;

	TRANSLATEBYME(ip->inverseV);
}

void RotateInvert(PyObject *this)
{

	PyInterpolatedPath *ip = (PyInterpolatedPath *)this;

        if (ip->inverseS > 0)
        	ROTATEBYME(ip->inverseV, ip->inverseS);
}

void PathApply(PyObject *this)
{
        PyInterpolatedPath *ip = (PyInterpolatedPath *)this;

        TRANSLATEBYME(ip->forwardV);
}

void RotateApply(PyObject *this)
{
        PyInterpolatedPath *ip = (PyInterpolatedPath *)this;

        if (ip->forwardS > 0)
                ROTATEBYME(ip->forwardV, ip->forwardS);

}

int C_PyInterpolatedPath_GetBaseTime(PyObject *self) {

	return ((PyInterpolatedPath *)self)->basePos;
	
}

PyObject *PyInterpolatedPath_GetBaseTime(PyObject *self, PyObject *args) {

	int baseTime = C_PyInterpolatedPath_GetBaseTime(self);
	
	return PyInt_FromLong((long)baseTime);
}

void C_PyInterpolatedPath_InvertByMe(PyObject *this)
{
	PyInterpolatedPath *ip = (PyInterpolatedPath *)this;
	ip->invertFunction(this);
}

void C_PyInterpolatedPath_TransformByMe(PyObject *this)
{
        PyInterpolatedPath *ip = (PyInterpolatedPath *)this;
        ip->transformFunction(this);
}

#define IP_INITED	(1<<0)
#define IP_CALLED 	(1<<1)
static void C_PyInterpolatedPath_GenerateTransform(PyObject *self, 
                PyObject *status, int curPos, int gen_forward, int gen_reverse,
                int apply) {
        PyInterpolatedPath *ip = (PyInterpolatedPath *)self;
        PyIPStatus *s = (PyIPStatus *)status;
	PyObject *translate1, *translate2;
	int pointIndex1, pointIndex2;

        if (!(ip->is_inited)) {
                ip->basePos = curPos;
                ip->is_inited = 1;
        }
                
        if (curPos - ip->basePos - s->delay >= ip->maxPos && 
                        s->doCallback != 0) {
                translate1 = PyList_GetItem(ip->pathPoints, 
							ip->maxIndex - 1);
                ip->interpFunction(self, translate1, translate1, 0, 1, 0,
                                gen_forward, gen_reverse, apply);
                if (!(s->is_called)) {
                        s->is_called = 1;
			Py_INCREF(self);
			if (s->userPointer == NULL) {
				Py_INCREF(Py_None);
				AppendMessage(std_queue, 
					MESSAGE_INTERPOLATED_PATH_COMPLETE, 
					self, Py_None);
			} else {
				Py_INCREF(s->userPointer);
				AppendMessage(std_queue, 
					MESSAGE_INTERPOLATED_PATH_COMPLETE, 
					self, s->userPointer);
			}
                }
                return;
        } else {
                s->is_called = 0;
        }

	
        curPos -= ip->basePos + s->delay;
        if (curPos > 0) {
		if (curPos - (curPos % ip->maxPos) > 0)
			s->delay += ip->maxPos;
                curPos %= ip->maxPos;
        } else {
                translate1 = PyList_GetItem(ip->pathPoints, 0);
		ip->interpFunction(self, translate1, translate1, 0, 1, 0,
                                gen_forward, gen_reverse, apply);
                return;
        }

	//printf("%d %d\n", curPos, s->nextInterval);
	if (s->genIntervals == 0)
		goto notest;
	if (curPos >= s->nextInterval)
	{
		if (curPos + s->intervalSize > ip->maxPos && s->wrap == 1)
			goto notest;
	
		s->wrap = 0;
domessage:	
		Py_INCREF(self);
		Py_INCREF(Py_None);
		AppendMessage(std_queue, MESSAGE_INTERPOLATED_PATH_INTERVAL, 
				self, Py_None);

		s->nextInterval += s->intervalSize;
		if (s->nextInterval > ip->maxPos) {
			s->nextInterval -= ip->maxPos;
			s->wrap = 1;
			goto notest;
		}
	} else if (curPos < s->intervalSize && s->wrap == 0 &&
			s->nextInterval + s->intervalSize > ip->maxPos) {

		goto domessage;

	}
		
notest:

	if (curPos < s->interpPointA) {
                IL_ITEM(ip->pathIndexes, 0, &(s->interpPointB));
                s->curIndex = -1;
                s->interpPointA = 0;
	}

	while (curPos >= s->interpPointB) {
		s->interpPointA = s->interpPointB;
		if (++(s->curIndex) == ip->maxIndex - 1) {
			s->interpPointB = ip->maxPos;
                } else {
                        IL_ITEM(ip->pathIndexes, s->curIndex + 1, 
                                        &(s->interpPointB));
		}
	}
	
	if (s->curIndex == -1) {
		pointIndex1 = pointIndex2 = 0;
	} else if (s->curIndex == ip->maxIndex - 1) {
		pointIndex1 = pointIndex2 = s->curIndex;
	} else {
		pointIndex1 = s->curIndex;
		pointIndex2 = pointIndex1 + 1;
	}

	translate2 = PyList_GetItem(ip->pathPoints, pointIndex2);
	translate1 = PyList_GetItem(ip->pathPoints, pointIndex1);

	ip->interpFunction(self, translate1, translate2, s->interpPointA, 
                        s->interpPointB, curPos, gen_forward, gen_reverse,
                        apply);

        

}

static int C_PyInterpolatedPath_RegisterSize(PyObject *self) {
        return sizeof(PyIPStatus);
}

static PyObject *PyInterpolatedPath_SetNotificationInterval(PyObject *self,
		PyObject *args) {

	PyInterpolatedPath *ip = (PyInterpolatedPath *)self;
	int handle, interval, offset;
	PyIPStatus *s;
	
	if (!PyArg_ParseTuple(args, "iii", &handle, &offset, &interval)) {
		return NULL;
	}

	s = (PyIPStatus *)handle;
	s->nextInterval = offset;
	s->intervalSize = interval;
	s->genIntervals = 1;
	
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *PyInterpolatedPath_SetIterationScheme(PyObject *self, 
		PyObject *args) {

	int scheme;
	int handle;
	PyIPStatus *s;
	PyInterpolatedPath *ip = (PyInterpolatedPath *)self;
	
	if (!PyArg_ParseTuple(args, "ii", &handle, &scheme)) {
		return NULL;
	}

	s = (PyIPStatus *)handle;
	
	if (scheme == 0)
	{
		s->doCallback = 0;
	}
	else
	{
		s->doCallback = 1;
		if (s->userPointer == NULL)
			s->is_called = 1;
		else
			s->is_called = 0;
	}

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *PyInterpolatedPath_SetDelay(PyObject *self, PyObject *args)
{
	int handle;
	PyIPStatus *s;
	int delay;

	if (!PyArg_ParseTuple(args, "ii", &handle, &delay)) {
		return NULL;
	}

	s = (PyIPStatus *)handle;

	s->delay = delay;

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *PyInterpolatedPath_SetCallback(PyObject *self, PyObject *args)
{

	int handle;
	PyObject *callback;
	PyIPStatus *s;
	
	if (!PyArg_ParseTuple(args, "iO", &handle, &callback)) {
		printf("PROBLEMO!!\n");
		return NULL;
	}

	s = (PyIPStatus *)handle;

	if (s->userPointer != NULL) {
		Py_DECREF(s->userPointer);
	}

	if (callback != Py_None) {
		Py_INCREF(callback);
		s->userPointer = callback;
		s->doCallback = 1;
		s->is_called = 0;
	} else {
		s->userPointer = NULL;
		s->doCallback = 0;
	}

	Py_INCREF(Py_None);
	return Py_None;
}	

static void *C_PyInterpolatedPath_Register(PyObject *self, PyObject *status, 
                PyObject *args) {
        
        PyInterpolatedPath *ip = (PyInterpolatedPath *)self;
        PyIPStatus *s = (PyIPStatus *)status;

        s->doCallback = 0;
	s->userPointer = NULL;
        s->delay = 0;

        if (args != NULL) {
                if (!PyArg_ParseTuple(args, "|iO", &s->delay, &s->userPointer))
		{
			printf("In Interpolated path Registration function:\n");
			PyErr_Print();
			return NULL;
		}

                if (s->userPointer != NULL) {
                        Py_INCREF(s->userPointer);
			/*
			 * Automatically set callback to 1 if user registers 
			 * callback stuff
			 */
			s->doCallback = 1;
                }


        }        
        
        s->curIndex = 0;
        s->is_called = 0;
        s->interpPointA = 0;
	s->nextInterval = ip->maxPos + 1;
	s->genIntervals = 0;
	s->wrap = 0;
	
        IL_ITEM(ip->pathIndexes, 1, &(s->interpPointB));

	return (void *)s;
}        

static void C_PyInterpolatedPath_DeRegister(PyObject *self, PyObject *status)
{

        PyIPStatus *s = (PyIPStatus *)status;

        
        if (s->userPointer != NULL) {
                Py_DECREF(s->userPointer);
        }

}

static PyObject *PyInterpolatedPath_SetMaxIndex(PyObject *self, PyObject *args) {
	
	int maxIndex;
	PyInterpolatedPath *ip = (PyInterpolatedPath *)self;

	if (!PyArg_ParseTuple(args, "i", &maxIndex)) {
		return NULL;
	}

	if (maxIndex < 2)
	{
		PyErr_SetString(PyExc_RuntimeError, 
			"max index of an interpolated path must be > 2");
		return NULL;
	}

	ip->maxPos = maxIndex;
	
	Py_INCREF(Py_None);
	return Py_None;
}

static void C_PyInterpolatedPath_SetType(PyObject *self, int type) {

        PyInterpolatedPath *ip = (PyInterpolatedPath *)self;
        
        if (type == BEZIER_PATH)
        {
                ip->interpFunction = BezierPathInterp;
                ip->invertFunction = PathInvert;
                ip->transformFunction = PathApply;
        }
        else if (type == LINEAR_PATH)
        {
                ip->interpFunction = LinearPathInterp;
                ip->invertFunction = PathInvert;
                ip->transformFunction = PathApply;
        }
        else if (type == LINEAR_ROTATE)
       	{
                ip->interpFunction = LinearRotateInterp;
                ip->invertFunction = RotateInvert;
                ip->transformFunction = RotateApply;
        }
        
}

static PyObject *PyInterpolatedPath_SetType(PyObject *self, PyObject *args) {

        int type;

        if (!PyArg_ParseTuple(args, "i", &type))
                return NULL;
        
        C_PyInterpolatedPath_SetType(self, type);

        Py_INCREF(Py_None);
        return Py_None;

}

static PyMethodDef PyInterpolatedPath_Methods[] = {
	{"append", (PyCFunction)PyInterpolatedPath_Append, METH_VARARGS, 
		"Appends a path point and path index to the path"},
	{"setMaxIndex", (PyCFunction)PyInterpolatedPath_SetMaxIndex, 
                METH_VARARGS, "Set the maximum index of the interpolated path"},
        {"setType", (PyCFunction)PyInterpolatedPath_SetType,
                METH_VARARGS, "Set the interpolation type"},
	{"getBaseTime", (PyCFunction)PyInterpolatedPath_GetBaseTime,
		METH_VARARGS, "Get the current internal base time"},
	{"setCallback", (PyCFunction)PyInterpolatedPath_SetCallback,
		METH_VARARGS, "Set the interpolated path's callback"},
	{"setIterationScheme", 
		(PyCFunction)PyInterpolatedPath_SetIterationScheme,
		METH_VARARGS, "Set the interpolated path's iteration scheme"},
	{"setNotificationInterval",
		(PyCFunction)PyInterpolatedPath_SetNotificationInterval,
		METH_VARARGS, 
		"Set the interpolate path's notification interval"},
	{"setDelay", (PyCFunction)PyInterpolatedPath_SetDelay, METH_VARARGS,
		"Set the interpolation path's delay"},
	{NULL, NULL, 0, NULL}
};

static PyObject *PyInterpolatedPath_GetAttr(PyObject *self, char *name) {
	
	return Py_FindMethod(PyInterpolatedPath_Methods, self, name);
	
}

static PySequenceMethods PyInterpolatedPath_sequenceMethods = {
        0, /* length */
        0, /* concat */
        0, /* repeat */
        0, /* item */
        0, /* slice */
        0, /* ass_item */
        0, /* ass_slice */
        0, /* contains */
        0, /* inplace_concat */
        0 /* inplace_repeat */
};

static PyTypeObject PyInterpolatedPathType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyInterpolatedPath",
        sizeof(PyInterpolatedPath),
        0,
        PyInterpolatedPath_Free,
        0, /* print */
        PyInterpolatedPath_GetAttr, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        &PyInterpolatedPath_sequenceMethods, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
