/*
 * Python object representing a GL Material
 */


#include "TograIncludes.h"
staticforward PyTypeObject PyMaterialType;

static void C_PyMaterial_SetMaterial(PyObject *material, int time);
static void C_PyMaterial_UnsetMaterial(PyObject *material, int time);

static IMaterial PyMaterial_IMaterial = {C_PyMaterial_SetMaterial, 
                C_PyMaterial_UnsetMaterial};
static IC PyMaterial_Interfaces = {NULL, NULL, NULL, NULL, NULL, NULL, 
        &PyMaterial_IMaterial};

PyMaterial *C_PyMaterial_New() {

        PyMaterial *m = (PyMaterial *)PyObject_New(PyMaterial, &PyMaterialType);
        int i;
        
        D_ALLOC(m, "Material");

        m->interfaces = &PyMaterial_Interfaces;
        
        for (i = 0; i < 4; i++) {
                m->ambient_colour[i] = 1;
                m->specular_colour[i] = 0;
                m->emissive_colour[i] = 0;
        }

	m->emissive_colour[3] = 1;
	m->specular_colour[3] = 1;
	
        m->shininess = 0;
        
        m->texture = NULL;

        return m;
}

void C_PyMaterial_SetAmbientColour(PyMaterial *m, float r, float g, float b,
                float a) {
	m->ambient_colour[0] = r;
        m->ambient_colour[1] = g;
        m->ambient_colour[2] = b;
        m->ambient_colour[3] = a;
}

void C_PyMaterial_SetEmissiveColour(PyMaterial *m, float r, float g, float b,
                float a) {
        m->emissive_colour[0] = r;
        m->emissive_colour[1] = g;
        m->emissive_colour[2] = b;
        m->emissive_colour[3] = a;
}


void C_PyMaterial_SetSpecularColour(PyMaterial *m, float r, float g, float b,
                float a) {
        m->specular_colour[0] = r;
        m->specular_colour[1] = g;
        m->specular_colour[2] = b;
        m->specular_colour[3] = a;
}        

void C_PyMaterial_SetShininess(PyMaterial *m, float s) {
        m->shininess = s;
}

PyObject *PyMaterial_New(PyObject *self, PyObject *args, PyObject *kwords) {

        PyMaterial *m;
        PyObject *amb = NULL, *spec = NULL, *em = NULL;
        float shin = 0;
        float r, g, b, a;
        
        char *kwlist[] = {"ambient", "specular", "shininess", "emissive", NULL};
        
        if (!PyArg_ParseTupleAndKeywords(args, kwords, "|OOfO", kwlist, &amb,
                                &spec, &shin, &em)) {
                return NULL;
        }
                
        m = C_PyMaterial_New();

        if (amb != NULL)
        {
                r = (float)PyFloat_AsDouble(PyTuple_GetItem(amb, 0));
                g = (float)PyFloat_AsDouble(PyTuple_GetItem(amb, 1));
                b = (float)PyFloat_AsDouble(PyTuple_GetItem(amb, 2));
                a = (float)PyFloat_AsDouble(PyTuple_GetItem(amb, 3));
        
                C_PyMaterial_SetAmbientColour(m, r, g, b, a);
        }

        if (spec != NULL)
        {
                r = (float)PyFloat_AsDouble(PyTuple_GetItem(spec, 0));
                g = (float)PyFloat_AsDouble(PyTuple_GetItem(spec, 1));
                b = (float)PyFloat_AsDouble(PyTuple_GetItem(spec, 2));
                a = (float)PyFloat_AsDouble(PyTuple_GetItem(spec, 3));
        
                C_PyMaterial_SetSpecularColour(m, r, g, b, a);
        }                        

        if (em != NULL)
        {
                r = (float)PyFloat_AsDouble(PyTuple_GetItem(em, 0));
                g = (float)PyFloat_AsDouble(PyTuple_GetItem(em, 1));
                b = (float)PyFloat_AsDouble(PyTuple_GetItem(em, 2));
                a = (float)PyFloat_AsDouble(PyTuple_GetItem(em, 3));
        
                C_PyMaterial_SetEmissiveColour(m, r, g, b, a);
        }                        
	
        C_PyMaterial_SetShininess(m, shin);
        
        return (PyObject *)(m);
}

static void PyMaterial_Free(PyObject *material) {

        D_FREE(material, "Material");
        
}

#define PFLOATPARM(s,t)	\
	TPRINTF("%s (%2.2f,%2.2f,%2.2f,%2.2f)\n", s, t[0], t[1], t[2], t[3]);

inline void _C_PyMaterial_SetMaterial(PyMaterial *m, int time) {

	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, m->ambient_colour);
        glMaterialfv(GL_FRONT, GL_SPECULAR, m->specular_colour);
        glMaterialf(GL_FRONT, GL_SHININESS, m->shininess);
        glMaterialfv(GL_FRONT, GL_EMISSION, m->emissive_colour);

#if DEBUG_VERBOSE
	PFLOATPARM("GL_AMBIENT_AND_DIFFUSE", m->ambient_colour);
	PFLOATPARM("GL_SPECULAR", m->specular_colour);
	TPRINTF("GL_SHININESS %2.2f\n", m->shininess);
	PFLOATPARM("GL_SHININESS", m->emissive_colour);
#endif

    	if (m->texture != NULL) {
		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
		TEXTURE_SET(m->texture, time);
	} else {
		glDisable(GL_TEXTURE_2D);
	}
}

extern PyMaterial *current_colour;

static void C_PyMaterial_SetMaterial(PyObject *material, int time) {

	PyMaterial *m = (PyMaterial *)material;
       
	m->old_colour = current_colour;
	current_colour = m;
	
	_C_PyMaterial_SetMaterial(current_colour, time);
}

static void C_PyMaterial_UnsetMaterial(PyObject *material, int time) {
	
	PyMaterial *m = (PyMaterial *)material;
	current_colour = m->old_colour;
	
	_C_PyMaterial_SetMaterial(current_colour, time);
}

static void C_PyMaterial_SetTexture(PyObject *self, PyObject *texture) {

        PyMaterial *m = (PyMaterial *)self;

	if (m->texture != NULL) {
		Py_DECREF(m->texture);
	}

	if (texture == Py_None) {
		m->texture = NULL;
		return;
	}
	m->texture = texture;
	Py_INCREF(texture);
}

static PyObject *PyMaterial_SetTexture(PyObject *self, PyObject *args) {

        PyObject *texture;
        
        if (!PyArg_ParseTuple(args, "O", &texture)) {
                return NULL;
        }

	
	C_PyMaterial_SetTexture(self, texture);
	
        Py_INCREF(Py_None);
        return Py_None;
        
}

static PyMethodDef PyMaterial_methods[] = {
        {"setTexture", (PyCFunction)PyMaterial_SetTexture, METH_VARARGS,
                "set the texture object for this material"},
        {NULL, NULL, 0, NULL}
};

static PyObject *PyMaterial_GetAttr(PyObject *self, char *name) {

        return Py_FindMethod(PyMaterial_methods, self, name);

}
        
static PyTypeObject PyMaterialType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyContainer",
        sizeof(PyMaterial),
        0,
        PyMaterial_Free,
        0, /* print */
        PyMaterial_GetAttr, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        0, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
