/*
 * Copyright 2003, Shane Stephens.
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

TOGRA_DEFINE_MIXIN(ODEMixin)

static void ODEnearCallback(void *data, dGeomID o1, dGeomID o2);

void *C_PyODEMixin_New(void) {

	PyODEMixin *om = PyObject_New(PyODEMixin, &PyODEMixinType);
	D_ALLOC(om, "ODEMixin");

	om->interfaces = &PyODEMixin_Interfaces;

	return (void *)om;

}

PyObject *PyODEMixin_New(PyObject *self, PyObject *args) {

	return C_PyODEMixin_New();

}

void C_PyODEMixin_Init(PyObject *self, int time) {
	
	PyODEMixin *om = (PyODEMixin *)self;
	om->lastTime = time;

	om->world = dWorldCreate();
	om->space = dHashSpaceCreate(0);
	om->contactGroup = dJointGroupCreate(0);
      
	dWorldSetGravity(om->world, 0, -4, 0);
	
}

#define MINTIME		200

void C_PyODEMixin_RenderStep(PyObject *self, int time, int is_paused,
		int pausedTime) {

	long timeDiff;
	
	PyODEMixin *om = (PyODEMixin *)self;
	
	if (is_paused)
		return;
	
	timeDiff = time - om->lastTime;
	om->lastTime = time;
	
	dSpaceCollide(om->space, self, *ODEnearCallback);
	while (timeDiff > MINTIME) {
		dWorldStep(om->world, MINTIME / 1000.0);
		timeDiff -= MINTIME;
	}
	dWorldStep(om->world, timeDiff / 1000.0);
	dJointGroupEmpty(om->contactGroup);

}

void C_PyODEMixin_Destroy(PyObject *self, int time) {
	
	return;
}

#define MAX_CONTACTS	4

static void ODEnearCallback(void *data, dGeomID o1, dGeomID o2)
{

	PyODEMixin *om = (PyODEMixin *)data;
	
	dBodyID b1, b2;
	int i;
	int numc;

	b1 = dGeomGetBody(o1);
	b2 = dGeomGetBody(o2);
	if (b1 && b2 && dAreConnectedExcluding(b1, b2, dJointTypeContact))
		return;
	
	dContact contact[MAX_CONTACTS];
	for (i = 0; i < MAX_CONTACTS; i++)
	{
		contact[i].surface.mode = 0; //dContactBounce; // | dContactSoftCFM;
		contact[i].surface.mu = dInfinity;
		contact[i].surface.mu2 = 0;
		contact[i].surface.bounce = 1;
		contact[i].surface.bounce_vel = 0.1;
		contact[i].surface.soft_cfm = 0.01;
	}

	if ((numc = dCollide(o1, o2, MAX_CONTACTS, &contact[0].geom,
				sizeof(dContact)))) {
		dMatrix3 RI;
		dRSetIdentity(RI);
		//const dReal ss[3] = {0.02, 0.02, 0.02};
		for (i = 0; i < numc; i++) {
			dJointID c = dJointCreateContact(om->world, 
					om->contactGroup, contact + i);
			dJointAttach(c, b1, b2);
		}
	}
}

static void PyODEMixin_Free(PyObject *self) {

	D_FREE(self, "ODEMixin");

}

static void C_PyODEMixin_SetGravity(PyObject *self, PyObject *grav) {

	PyODEMixin *om = (PyODEMixin *)self;

	PyVector *g = (PyVector *)grav;
	
	dWorldSetGravity(om->world, g->vector[0], g->vector[1], g->vector[2]);
}

static PyObject *PyODEMixin_SetGravity(PyObject *self, PyObject *args)
{

	PyObject *grav;
	
	if (!PyArg_ParseTuple(args, "O", &grav)) {
		return NULL;
	}

	C_PyODEMixin_SetGravity(self, grav);

	Py_INCREF(Py_None);
	return Py_None;
}


static PyMethodDef PyODEMixin_methods[] = {
	{"setGravity", (PyCFunction)PyODEMixin_SetGravity, METH_VARARGS,
		"set the gravity of this ODE object"},
	{NULL, NULL, 0, NULL}
};

static PyObject *PyODEMixin_GetAttr(PyObject *self, char *name) {
	return Py_FindMethod(PyODEMixin_methods, self, name);
}

TOGRA_DEFINE_GETATTR_OBJECT(ODEMixin)
