/*
 * Copyright 2003, Shane Stephens.
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

/*
 * Requires:    A render tree to reflect (IRender)
 *              A reflection transformation (ITransform)
 *              A surface to present the reflection on (IRender)
 * 
 * Provides:    A reflection (sort of, maybe, kinda)
 */

#include "TograIncludes.h"

staticforward PyTypeObject PyReflectType;

void C_PyReflect_Render(PyObject *self, int time, int picking);
int C_PyReflect_Compile(PyObject *self) {return FALSE;}

static IRender PyReflect_IRender = {C_PyReflect_Render, C_PyReflect_Compile};
static IC PyReflect_Interfaces = {&PyReflect_IRender};

PyObject *C_PyReflect_New(PyObject *face, PyObject *transform, PyObject *tree) {

        PyReflect *r = PyObject_New(PyReflect, &PyReflectType);

        D_ALLOC(r, "Reflect");
        
        r->reflectFace = face;
        r->reflectTransform = transform;
        r->reflectedTree = tree;
        r->interfaces = &PyReflect_Interfaces;
	r->path = NULL;
	r->status = NULL;

        Py_INCREF(face);
        Py_INCREF(transform);
        Py_INCREF(tree);

        return (PyObject *)r;

}

void C_PyReflect_Free(PyObject *self) {

        PyReflect *r = (PyReflect *)self;

        D_FREE(r, "Reflect");
        
        Py_DECREF(r->reflectFace);
        Py_DECREF(r->reflectTransform);
        Py_DECREF(r->reflectedTree);

        PyObject_Del(self);

}

PyObject *PyReflect_New(PyObject *self, PyObject *args) {

        PyObject *face, *transform, *tree;

        if (!PyArg_ParseTuple(args, "OOO", &face, &transform, &tree)) {
                return NULL;
        }

        return C_PyReflect_New(face, transform, tree);

}

PyObject *C_PyMatrix_FromScale(double *scale);
void C_PyRenderControl_SetupLightingTransform(void);
void C_PyRenderControl_DoLighting(int time);

void C_PyReflect_Render(PyObject *self, int time, int picking) {

        PyReflect *r = (PyReflect *)self;
        PyObject *rmat;
        double scale[] = {1.0, 1.0, -1.0};

        /*
         * 1. Render the surface into the stencil buffer
         *
         * 2. Transform by the ITransform
         *
         * 3. Render the render tree
         *
         * 4. Pop the transformation
         */

        glEnable(GL_STENCIL_TEST);
       	glStencilMask(1);
	
        glClear(GL_STENCIL_BUFFER_BIT);
        //glClearStencil(1);
	
        glStencilFunc(GL_ALWAYS, 1, 1);
        glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);
        glColorMask(0,0,0,0);
        glDisable(GL_DEPTH_TEST);

        glPushMatrix();
        
	GENERATE(r->reflectTransform, NULL, time, 1, 0, 1);
	RENDER(r->reflectFace, time, picking);
	glPopMatrix();        

        glEnable(GL_DEPTH_TEST);
        glColorMask(1,1,1,1);
        glStencilFunc(GL_EQUAL,1,1);
        glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

        glDisable(GL_CULL_FACE);
       
        rmat = C_PyMatrix_FromScale(scale);
	
	/*
	 * Do lighting
	 */
        glPushMatrix();
        C_PyRenderControl_SetupLightingTransform();
	GENERATE(r->reflectTransform, r->status, time, 1, 1, 1);
        GENERATE(rmat, NULL, 0, 0, 1, 1);
        INVERTBYME(r->reflectTransform);
	C_PyRenderControl_DoLighting(time);
	glPopMatrix();
	
	glPushMatrix();
        GENERATE(r->reflectTransform, r->status, time, 1, 1, 1);
        GENERATE(rmat, NULL, 0, 0, 1, 1);
        INVERTBYME(r->reflectTransform);

        RENDER(r->reflectedTree, time, picking);

        glPopMatrix();

	/*
	 * Restore lighting
	 */
	glPushMatrix();
        C_PyRenderControl_SetupLightingTransform();
	C_PyRenderControl_DoLighting(time);
	glPopMatrix();
	
        glEnable(GL_CULL_FACE);
        glCullFace(GL_BACK);
        glDisable(GL_STENCIL_TEST);
}

void C_PyReflect_SetPath(PyObject *self, PyObject *path, int delay,
			PyObject *callback, PyObject *user1, PyObject *user2)
{
        void *r;
	int size;

	PyReflect *this = (PyReflect *)self;

	if (this->status != NULL)
	{
		DEREGISTER(this->path, this->status);
		free(this->status);
		Py_DECREF(this->path);
	}

	Py_INCREF(path);
	this->path = path;

        size = REGISTERSIZE(path);
        r = malloc(size);
        //REGISTER(path, r, delay, callback, user1, user2);
        
        this->status = r;

}

PyObject *PyReflect_SetPath(PyObject *this, PyObject *args)
{

        PyObject *callback = NULL, *user1 = NULL, *user2 = NULL, *path;
        int delay = 0;
        
        if (!PyArg_ParseTuple(args, "O|iOOO", &path, &delay, &callback, 
				&user1, &user2)) {
                return NULL;
        }

	C_PyReflect_SetPath(this, path, delay, callback, user1, user2);
	
	Py_INCREF(Py_None);
	return Py_None;
}

static PyMethodDef PyReflect_methods[] = {
        {"setPath", (PyCFunction)PyReflect_SetPath, METH_VARARGS,
                "Set the reflection surface's path"},
        {NULL, NULL, 0, NULL}
};

static PyObject *PyReflect_GetAttr(PyObject *self, char *name) {

        return Py_FindMethod(PyReflect_methods, self, name);

}

static PyTypeObject PyReflectType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyReflect",
        sizeof(PyReflect),
        0,
        C_PyReflect_Free,
        0, /* print */
        PyReflect_GetAttr, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        0, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};

