#include "gsttograsink.h"
#include <gst/video/video.h>

#define TOGRA_SINK_DEFAULT_SIZE 256

static GstElementDetails gst_tograsink_details =
GST_ELEMENT_DETAILS (
    "Video Sink",
    "Sink/Video",
    "Video Output to Togra texture",
    "Jan Schmidt <thaytan@mad.scientist.com>");

static GstStaticPadTemplate gst_tograsink_sink_template =
    GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ( GST_VIDEO_CAPS_RGB )
);

enum
{
  ARG_0,
  ARG_WIDTH,
  ARG_HEIGHT
};

static void gst_tograsink_finalize (GObject *object);

static GstElementClass *parent_class = NULL;

static void
gst_tograsink_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstTograSink *tograsink;

  g_return_if_fail (GST_IS_TOGRASINK (object));
  tograsink = GST_TOGRASINK (object);

  switch (prop_id) {
    case ARG_WIDTH:
      tograsink->set_width = g_value_get_int (value);
      break;
    case ARG_HEIGHT:
      tograsink->set_height = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_tograsink_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstTograSink *tograsink;

  g_return_if_fail (GST_IS_TOGRASINK (object));
  tograsink = GST_TOGRASINK (object);

  switch (prop_id) {
    case ARG_WIDTH:
      g_value_set_int (value, tograsink->set_width);
      break;
    case ARG_HEIGHT:
      g_value_set_int (value, tograsink->set_height);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_tograsink_set_clock (GstElement * element, GstClock * clock)
{
  GstTograSink *tograsink;

  tograsink = GST_TOGRASINK (element);

  tograsink->clock = clock;
}

static void
gst_tograsink_chain (GstPad *pad, GstData *data)
{
	GstBuffer 	  *buf = NULL;
	GstTograSink 	*tograsink;
  GstClockTime  buf_time;
  
	g_return_if_fail (GST_IS_PAD (pad));
	g_return_if_fail (data != NULL);

	tograsink = GST_TOGRASINK (gst_pad_get_parent (pad));
	if (GST_IS_EVENT (data)) {
    		gst_pad_event_default (pad, GST_EVENT (data));
    		return;
  }

  buf = GST_BUFFER (data);
  if (GST_BUFFER_TIMESTAMP_IS_VALID (buf)) {
    tograsink->time = GST_BUFFER_TIMESTAMP (buf);
  }

  buf_time = tograsink->time;
  
  if (tograsink->clock != NULL) {
    gst_element_wait (GST_ELEMENT (tograsink), buf_time);
  }

  if (GST_BUFFER_DURATION_IS_VALID (buf)) {
    tograsink->time += GST_BUFFER_DURATION (buf);
  }
  else if (tograsink->framerate > 0) {
    tograsink->time += GST_SECOND / tograsink->framerate;
  }

  /* Lock the mutex and update the output buffer */
  g_mutex_lock (tograsink->out_mutex);
  
  if (tograsink->waiting_buffer) {
    gst_buffer_unref (tograsink->waiting_buffer);
  }
  tograsink->waiting_buffer = buf;
  
  tograsink->waiting_height = tograsink->cur_height;
  tograsink->waiting_width = tograsink->cur_width;
  tograsink->waiting_time = buf_time;

  if ((tograsink->cur_height != tograsink->set_height) ||
     (tograsink->cur_width != tograsink->set_width)) {
    gst_pad_renegotiate (tograsink->sink_pad);
    //g_print ("Renegotiated size to w,h %d, %d\n", tograsink->cur_width, tograsink->cur_height);
  }

  g_mutex_unlock (tograsink->out_mutex);    
}

GstTograSinkBuffer *
gst_tograsink_get_buffer (GstTograSink *tograsink)
{
  /* Check the pointer before bothering to lock */
  if (!tograsink->waiting_buffer)
    return NULL;    

  g_mutex_lock (tograsink->out_mutex);
  if (tograsink->curout_buffer)
    gst_buffer_unref (tograsink->curout_buffer);
    
  tograsink->curout_buffer = tograsink->waiting_buffer;
  tograsink->waiting_buffer = NULL;
  
  tograsink->out_buffer.data = GST_BUFFER_DATA (tograsink->curout_buffer);
  tograsink->out_buffer.time = tograsink->waiting_time;
  tograsink->out_buffer.width = tograsink->waiting_width;
  tograsink->out_buffer.height = tograsink->waiting_height;
	
  g_mutex_unlock (tograsink->out_mutex);

  return &(tograsink->out_buffer);
}

static GstCaps *
gst_tograsink_getcaps (GstPad * pad)
{
  GstTograSink *tograsink;
  GstCaps *caps = gst_caps_copy (gst_pad_get_pad_template_caps (pad));

  tograsink = GST_TOGRASINK (gst_pad_get_parent (pad));

#if 0
  gst_caps_set_simple (caps, "width", GST_TYPE_INT_RANGE, 16, 
	tograsink->set_width, "height", GST_TYPE_INT_RANGE, 16, 
	tograsink->set_height, NULL);
#else
  gst_caps_set_simple (caps, "width", G_TYPE_INT,
	tograsink->set_width, "height", G_TYPE_INT,
	tograsink->set_height, NULL);
#endif

  return caps;
}

static GstPadLinkReturn
gst_tograsink_sink_link (GstPad * pad, const GstCaps * caps)
{
  GstTograSink *tograsink;
  GstStructure *structure;
  gboolean ret;

  tograsink = GST_TOGRASINK (gst_pad_get_parent (pad));
  structure = gst_caps_get_structure (caps, 0);
  ret = gst_structure_get_int (structure, "width",
      &(tograsink->cur_width));
  ret &= gst_structure_get_int (structure, "height",
      &(tograsink->cur_height));
  ret &= gst_structure_get_double (structure, "framerate",
      &tograsink->framerate);

  if (tograsink->cur_width > tograsink->set_width)
    return GST_PAD_LINK_REFUSED;
	  
  if (tograsink->cur_height > tograsink->set_height)
    return GST_PAD_LINK_REFUSED;

  if (!ret)
    return GST_PAD_LINK_REFUSED;

  return GST_PAD_LINK_OK;
}

static GstElementStateReturn
gst_tograsink_change_state (GstElement * element)
{
  GstTograSink *tograsink;

  tograsink = GST_TOGRASINK (element);

  switch (GST_STATE_TRANSITION (element)) {
    case GST_STATE_READY_TO_PAUSED:
    tograsink->time = 0; 
    break;
  }

  if (GST_ELEMENT_CLASS (parent_class)->change_state)
        return GST_ELEMENT_CLASS (parent_class)->change_state (element);

  return GST_STATE_SUCCESS;
}

static void
gst_tograsink_base_init (gpointer g_class)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (g_class);

  gst_element_class_set_details (element_class, &gst_tograsink_details);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&gst_tograsink_sink_template));
}

static void
gst_tograsink_class_init (GstTograSinkClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  g_object_class_install_property (gobject_class, ARG_WIDTH,
      g_param_spec_int ("width", "width", 
	      "The maximum width for the output texture", 
	      1, 4096, TOGRA_SINK_DEFAULT_SIZE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_HEIGHT,
      g_param_spec_int ("height", "height", 
		"The maximum height for the output texture",
          	1, 4096, TOGRA_SINK_DEFAULT_SIZE, G_PARAM_READWRITE));

  gobject_class->finalize = gst_tograsink_finalize;
  gobject_class->set_property = gst_tograsink_set_property;
  gobject_class->get_property = gst_tograsink_get_property;

  gstelement_class->change_state = gst_tograsink_change_state;
  gstelement_class->set_clock = gst_tograsink_set_clock;
}


static void
gst_tograsink_init (GstTograSink *tograsink)
{
  tograsink->sink_pad =
      gst_pad_new_from_template (gst_static_pad_template_get
      (&gst_tograsink_sink_template), "sink");

  gst_pad_set_chain_function (tograsink->sink_pad,
      gst_tograsink_chain);
  gst_pad_set_link_function (tograsink->sink_pad,
      gst_tograsink_sink_link);
  gst_pad_set_getcaps_function (tograsink->sink_pad,
      gst_tograsink_getcaps);

  gst_element_add_pad (GST_ELEMENT (tograsink), tograsink->sink_pad);

  tograsink->time = 0;
  tograsink->clock = NULL;
  tograsink->cur_width = -1;
  tograsink->cur_height = -1;
  tograsink->framerate = 0;
  tograsink->set_width = TOGRA_SINK_DEFAULT_SIZE;
  tograsink->set_height = TOGRA_SINK_DEFAULT_SIZE;
  
  tograsink->out_mutex = g_mutex_new();
  tograsink->waiting_buffer = NULL;
  tograsink->curout_buffer = NULL;
}

static void
gst_tograsink_finalize (GObject *object)
{
  GstTograSink *tograsink = GST_TOGRASINK(object);
  g_mutex_free (tograsink->out_mutex);

  if (tograsink->waiting_buffer) 
    gst_buffer_unref (tograsink->waiting_buffer);
  if (tograsink->curout_buffer) 
    gst_buffer_unref (tograsink->curout_buffer);
}

GType gst_tograsink_get_type (void)
{
	static GType tograsink_type = 0;
	if (!tograsink_type) {
    		static const GTypeInfo tograsink_info = {
      				sizeof (GstTograSinkClass),
      				gst_tograsink_base_init,
      				NULL,
      				(GClassInitFunc) gst_tograsink_class_init,
      				NULL,
      				NULL,
      				sizeof (GstTograSink),
      				0,
      				(GInstanceInitFunc) gst_tograsink_init,
    		};
		
    		tograsink_type =
        		g_type_register_static (GST_TYPE_ELEMENT, 
			"GstTograSink", &tograsink_info, 0);
	}

	return tograsink_type;
}

